#!/usr/bin/python
"""Map symbol exports/imports between a set of object files.

 This script will use the 'nm' utility to retrieve a list of
 symbols used and exported, generating a file that can be
 used to audit dependencies between object files, modules,
 and libraries.

 See http://www.asterisk.org for more information about
 the Asterisk project. Please do not directly contact
 any of the maintainers of this project for assistance;
 the project provides a web site, mailing lists and IRC
 channels for your use.

 This program is free software, distributed under the terms of
 the GNU General Public License Version 2. See the LICENSE file
 at the top of the source tree.

 Copyright (C) 2015, CFWare, LLC.
 Corey Farrell <git@cfware.com>
"""

import subprocess
import re
import os
import sys
import argparse


def remove_prefix(text, prefix):
    return text[text.startswith(prefix) and len(prefix):]


class DisplayProcessor(object):
    def __init__(self, args):
        self.prefix = args.prefix
        self.stripexten = args.stripexten
        self.color = args.color
        self.set_color(args.color)

    def set_color(self, color):
        if color:
            self.green = '\033[92m'
            self.end = '\033[0m'
        else:
            self.green = ''
            self.end = ''

    def reset_color(self):
        self.set_color(self.color)

    def nocolor_file(self, filename):
        filename = remove_prefix(filename, self.prefix)
        if self.stripexten:
            return os.path.splitext(filename)[0]
        return filename

    def color_file(self, filename):
        return "%s%s%s" % (self.green, self.nocolor_file(filename), self.end)

    def symbol(self, symbol):
        return "%s%s%s" % (self.green, symbol, self.end)


class ObjectResults(object):
    def __init__(self, filename, scanner):
        self.scanner = scanner
        if scanner.args.singleline:
            self.splitter = ' '
        else:
            self.splitter = "\n\t"
        self.outproc = self.scanner.outproc
        self.filename = filename
        # self.imports[symbol] = scanner.exports[sym]
        self.imports = {}
        # Dict: filename=[list of imported symbols]
        self.importfrom = {}
        # list of files we import from
        self.importfiles = []
        # Dict: exported symbol=[list of importers]
        self.exports = {}
        # list of files we export to
        self.exportto = []
        p = subprocess.Popen(['nm', '-g', filename], stdout=subprocess.PIPE)
        for line in iter(p.stdout.readline, ''):
            symbol = line[19:].rstrip('\n')
            if line[:19] == '                 U ':
                # import symbol, will point to ObjectResults once resolved
                self.imports[symbol] = None
            else:
                # export symbol
                self.exports[symbol] = []
                scanner.exports[symbol] = self

    def record_export(self, symbol, importer):
        if importer not in self.exports[symbol]:
            self.exports[symbol].append(importer)
        if importer not in self.exportto:
            self.exportto.append(importer)

    def record_import(self, symbol, exporter):
        self.imports[symbol] = exporter
        if exporter.filename not in self.importfrom:
            self.importfrom[exporter.filename] = []
        self.importfrom[exporter.filename].append(symbol)

    def resolve_import(self, sym):
        if sym not in self.scanner.exports:
            return

        export = self.scanner.exports[sym]
        export.record_export(sym, self)
        self.record_import(sym, export)

    def resolve_imports(self):
        for sym in self.imports.iterkeys():
            self.resolve_import(sym)

    def print_map(self, fd):
        self.print_innercore(fd, [])

    def print_imports(self, fd):
        for src in sorted(self.importfrom.iterkeys()):
            print >>fd, self.splitter.join(["%s:" % self.outproc.color_file(src)] +
                                    sorted(self.importfrom[src]))

    def print_exportto(self, fd):
        print >>fd, self.splitter.join(
            ["%s:" % self.outproc.color_file(self.filename)] +
            [self.outproc.nocolor_file(x.filename)
             for x in sorted(self.exportto, key=lambda x: x.filename)]
        )

    def print_exports(self, fd):
        for sym in sorted(self.exports.iterkeys()):
            importers = self.exports[sym]
            print >>fd, self.splitter.join(
                ["%s:" % self.outproc.symbol(sym)] +
                [self.outproc.nocolor_file(x.filename)
                 for x in sorted(importers, key=lambda x: x.filename)]
            )

    def print_innercore(self, fd, innercore):
        print >>fd, self.splitter.join(
            ["%s:" % self.outproc.color_file(self.filename)] +
            [self.outproc.nocolor_file(x)
             for x in sorted(self.importfrom.iterkeys()) if x not in innercore]
        )


class ObjectScanner(object):
    def __init__(self):
        self.objects = {}
        self.exports = {}

    def parse_command_line(self):
        defaultcolor = os.isatty(sys.stdout.fileno())
        parser = argparse.ArgumentParser()
        parser.add_argument('--objects', metavar='file', dest='filenames', type=str,
                            nargs='+',
                            help='List of object files to scan')
        parser.add_argument('--innercore', metavar='file', dest='innercore', type=str,
                            nargs='+',
                            help='List of object files to treat as part of a inner-core')
        parser.add_argument('--map', dest='print_map', action='store_const',
                            const=True, default=False,
                            help='Display a map of object file link dependencies')
        parser.add_argument('--imports', metavar='file.o', dest='imports', type=str,
                            help='List imported symbols')
        parser.add_argument('--exports', metavar='file.o', dest='exports', type=str,
                            help='List exported symbols')
        parser.add_argument('--find-symbol', metavar='symbol', dest='find_symbol',
                            type=str,
                            help='Find a symbol export')
        parser.add_argument('--color', dest='color', action='store_const', const=True,
                            default=defaultcolor,
                            help='Colorize output')
        parser.add_argument('--no-color', dest='color', action='store_const', const=False,
                            default=not defaultcolor,
                            help='Disable colorized output')
        parser.add_argument('--strip-ext', dest='stripexten', action='store_const',
                            const=True, default=False,
                            help='Strip object file extensions from output')
        parser.add_argument('--prefix', metavar='prefix', dest='prefix', type=str,
                            default='',
                            help='Strip prefix from filenames in output')
        parser.add_argument('--output-dir', metavar='output-dir', dest='output',
                            type=str, default=None,
                            help='Output all results to files in output-dir')
        parser.add_argument('--multi-line', dest='singleline', action='store_const',
                            const=False, default=True,
                            help='Split each target to a new line.')
        return parser.parse_args()

    def print_map(self, fd):
        for obj in sorted(self.objects.itervalues(), key=lambda x: x.filename):
            obj.print_map(fd)

    def print_exportto(self, fd):
        for obj in sorted(self.objects.itervalues(), key=lambda x: x.filename):
            obj.print_exportto(fd)

    def print_imports(self, filename, fd):
        print >>fd, "Imports for %s" % (self.outproc.color_file(filename))
        self.objects[filename].print_imports(fd)

    def print_exports(self, filename, fd):
        print >>fd, "Exports for %s" % (self.outproc.color_file(filename))
        self.objects[filename].print_exports(fd)

    def print_innercore(self, fd):
        print >>fd, "Inner-core:"
        for obj in sorted(self.objects.itervalues(), key=lambda x: x.filename):
            if obj.filename not in self.args.innercore:
                continue
            obj.print_innercore(fd, self.args.innercore)

        print >>fd, "\nOthers:"
        for obj in sorted(self.objects.itervalues(), key=lambda x: x.filename):
            if obj.filename in self.args.innercore:
                continue
            obj.print_innercore(fd, self.args.innercore)

    def produce_output_dir(self, outputdir):
        if os.path.lexists(outputdir):
            print "Output location already exists: %s" % outputdir
            exit(1)

        os.makedirs(outputdir)
        self.outproc.set_color(False)

        self.print_map(open(os.path.join(outputdir, 'usage_map'), 'w+'))
        self.print_exportto(open(os.path.join(outputdir, 'export_map'), 'w+'))

        if self.args.innercore:
            self.print_innercore(open(os.path.join(outputdir, 'innercore'), 'w+'))

        for obj in self.objects.itervalues():
            strippedfn = self.outproc.nocolor_file(obj.filename)
            path = os.path.join(outputdir, os.path.dirname(strippedfn))
            if not os.path.lexists(path):
                os.makedirs(path)

            noext = os.path.splitext(strippedfn)[0]
            obj.print_imports(open(os.path.join(outputdir, "%s.imports" % noext), 'w+'))
            obj.print_exports(open(os.path.join(outputdir, '%s.exports' % noext), 'w+'))

        self.outproc.reset_color()

    def run(self):
        args = self.args = self.parse_command_line()
        self.outproc = DisplayProcessor(args)

        if args.output and os.path.lexists(args.output):
            print "Output location already exists: %s" % args.output
            exit(1)

        for filename in args.filenames:
            if filename[1] == '/' or filename[1] == '.':
                print >>sys.stderr, "Filenames must not start with '/' or '.'"
                sys.exit(1)
            self.objects[filename] = ObjectResults(filename, self)

        for obj in self.objects.itervalues():
            obj.resolve_imports()

        if args.print_map:
            self.print_map(sys.stdout)

        if args.imports:
            self.print_imports(args.imports, sys.stdout)

        if args.exports:
            self.print_exports(args.exports, sys.stdout)

        if args.find_symbol in self.exports:
            print "Symbol %s is provided by %s" % (
                self.outproc.symbol(args.find_symbol),
                self.outproc.color_file(self.exports[args.find_symbol].filename)
            )

        if args.innercore and not args.output:
            self.print_innercore(sys.stdout)

        if args.output:
            self.produce_output_dir(args.output)


ObjectScanner().run()
